@@ -2,6 +2,7 @@
 
 namespace PHPStan\PhpDocParser\Parser;
 
+use Exception;
 use PHPStan\PhpDocParser\Ast\ConstExpr\ConstExprFloatNode;
 use PHPStan\PhpDocParser\Ast\ConstExpr\ConstExprIntegerNode;
 use PHPStan\PhpDocParser\Ast\ConstExpr\ConstExprStringNode;
@@ -11,17 +12,23 @@ use PHPStan\PhpDocParser\Ast\Type\ArrayShapeNode;
 use PHPStan\PhpDocParser\Ast\Type\ArrayTypeNode;
 use PHPStan\PhpDocParser\Ast\Type\CallableTypeNode;
 use PHPStan\PhpDocParser\Ast\Type\CallableTypeParameterNode;
+use PHPStan\PhpDocParser\Ast\Type\ConditionalTypeForParameterNode;
+use PHPStan\PhpDocParser\Ast\Type\ConditionalTypeNode;
 use PHPStan\PhpDocParser\Ast\Type\ConstTypeNode;
 use PHPStan\PhpDocParser\Ast\Type\GenericTypeNode;
 use PHPStan\PhpDocParser\Ast\Type\IdentifierTypeNode;
 use PHPStan\PhpDocParser\Ast\Type\IntersectionTypeNode;
 use PHPStan\PhpDocParser\Ast\Type\NullableTypeNode;
+use PHPStan\PhpDocParser\Ast\Type\OffsetAccessTypeNode;
 use PHPStan\PhpDocParser\Ast\Type\ThisTypeNode;
 use PHPStan\PhpDocParser\Ast\Type\TypeNode;
 use PHPStan\PhpDocParser\Ast\Type\UnionTypeNode;
 use PHPStan\PhpDocParser\Lexer\Lexer;
+use PHPUnit\Framework\TestCase;
+use function get_class;
+use const PHP_EOL;
 
-class TypeParserTest extends \PHPUnit\Framework\TestCase
+class TypeParserTest extends TestCase
 {
 
 	/** @var Lexer */
@@ -40,13 +47,11 @@ class TypeParserTest extends \PHPUnit\Framework\TestCase
 
 	/**
 	 * @dataProvider provideParseData
-	 * @param string              $input
-	 * @param TypeNode|\Exception $expectedResult
-	 * @param int                 $nextTokenType
+	 * @param TypeNode|Exception $expectedResult
 	 */
 	public function testParse(string $input, $expectedResult, int $nextTokenType = Lexer::TOKEN_END): void
 	{
-		if ($expectedResult instanceof \Exception) {
+		if ($expectedResult instanceof Exception) {
 			$this->expectException(get_class($expectedResult));
 			$this->expectExceptionMessage($expectedResult->getMessage());
 		}
@@ -146,6 +151,17 @@ class TypeParserTest extends \PHPUnit\Framework\TestCase
 					new IdentifierTypeNode('int'),
 				]),
 			],
+			[
+				'(' . PHP_EOL .
+				'  string' . PHP_EOL .
+				'  &' . PHP_EOL .
+				'  int' . PHP_EOL .
+				')',
+				new IntersectionTypeNode([
+					new IdentifierTypeNode('string'),
+					new IdentifierTypeNode('int'),
+				]),
+			],
 			[
 				'string & int & float',
 				new IntersectionTypeNode([
@@ -436,7 +452,7 @@ class TypeParserTest extends \PHPUnit\Framework\TestCase
 			],
 			[
 				'array{',
-				new \PHPStan\PhpDocParser\Parser\ParserException(
+				new ParserException(
 					'',
 					Lexer::TOKEN_END,
 					6,
@@ -445,7 +461,7 @@ class TypeParserTest extends \PHPUnit\Framework\TestCase
 			],
 			[
 				'array{a => int}',
-				new \PHPStan\PhpDocParser\Parser\ParserException(
+				new ParserException(
 					'=>',
 					Lexer::TOKEN_OTHER,
 					8,
@@ -732,6 +748,13 @@ class TypeParserTest extends \PHPUnit\Framework\TestCase
 				new IdentifierTypeNode('array'),
 				Lexer::TOKEN_OPEN_SQUARE_BRACKET,
 			],
+			[
+				'array[ int ]',
+				new OffsetAccessTypeNode(
+					new IdentifierTypeNode('array'),
+					new IdentifierTypeNode('int')
+				),
+			],
 			[
 				"?\t\xA009", // edge-case with \h
 				new NullableTypeNode(
@@ -1002,6 +1025,247 @@ class TypeParserTest extends \PHPUnit\Framework\TestCase
 				'int|array{}',
 				new UnionTypeNode([new IdentifierTypeNode('int'), new ArrayShapeNode([])]),
 			],
+			[
+				'callable(' . PHP_EOL .
+				'  Foo' . PHP_EOL .
+				'): void',
+				new CallableTypeNode(
+					new IdentifierTypeNode('callable'),
+					[
+						new CallableTypeParameterNode(new IdentifierTypeNode('Foo'), false, false, '', false),
+					],
+					new IdentifierTypeNode('void')
+				),
+			],
+			[
+				'callable(' . PHP_EOL .
+				'  Foo,' . PHP_EOL .
+				'  Bar' . PHP_EOL .
+				'): void',
+				new CallableTypeNode(
+					new IdentifierTypeNode('callable'),
+					[
+						new CallableTypeParameterNode(new IdentifierTypeNode('Foo'), false, false, '', false),
+						new CallableTypeParameterNode(new IdentifierTypeNode('Bar'), false, false, '', false),
+					],
+					new IdentifierTypeNode('void')
+				),
+			],
+			[
+				'callable(' . PHP_EOL .
+				'  Foo, Bar' . PHP_EOL .
+				'): void',
+				new CallableTypeNode(
+					new IdentifierTypeNode('callable'),
+					[
+						new CallableTypeParameterNode(new IdentifierTypeNode('Foo'), false, false, '', false),
+						new CallableTypeParameterNode(new IdentifierTypeNode('Bar'), false, false, '', false),
+					],
+					new IdentifierTypeNode('void')
+				),
+			],
+			[
+				'callable(' . PHP_EOL .
+				'  Foo,' . PHP_EOL .
+				'  callable(' . PHP_EOL .
+				'    Bar' . PHP_EOL .
+				'  ): void' . PHP_EOL .
+				'): void',
+				new CallableTypeNode(
+					new IdentifierTypeNode('callable'),
+					[
+						new CallableTypeParameterNode(new IdentifierTypeNode('Foo'), false, false, '', false),
+						new CallableTypeParameterNode(
+							new CallableTypeNode(
+								new IdentifierTypeNode('callable'),
+								[
+									new CallableTypeParameterNode(new IdentifierTypeNode('Bar'), false, false, '', false),
+								],
+								new IdentifierTypeNode('void')
+							),
+							false,
+							false,
+							'',
+							false
+						),
+					],
+					new IdentifierTypeNode('void')
+				),
+			],
+			[
+				'callable(' . PHP_EOL .
+				'  Foo,' . PHP_EOL .
+				'  callable(' . PHP_EOL .
+				'    Bar,' . PHP_EOL .
+				'  ): void' . PHP_EOL .
+				'): void',
+				new CallableTypeNode(
+					new IdentifierTypeNode('callable'),
+					[
+						new CallableTypeParameterNode(new IdentifierTypeNode('Foo'), false, false, '', false),
+						new CallableTypeParameterNode(
+							new CallableTypeNode(
+								new IdentifierTypeNode('callable'),
+								[
+									new CallableTypeParameterNode(new IdentifierTypeNode('Bar'), false, false, '', false),
+								],
+								new IdentifierTypeNode('void')
+							),
+							false,
+							false,
+							'',
+							false
+						),
+					],
+					new IdentifierTypeNode('void')
+				),
+			],
+			[
+				'(Foo is Bar ? never : int)',
+				new ConditionalTypeNode(
+					new IdentifierTypeNode('Foo'),
+					new IdentifierTypeNode('Bar'),
+					new IdentifierTypeNode('never'),
+					new IdentifierTypeNode('int'),
+					false
+				),
+			],
+			[
+				'(Foo is not Bar ? never : int)',
+				new ConditionalTypeNode(
+					new IdentifierTypeNode('Foo'),
+					new IdentifierTypeNode('Bar'),
+					new IdentifierTypeNode('never'),
+					new IdentifierTypeNode('int'),
+					true
+				),
+			],
+			[
+				'(T is self::TYPE_STRING ? string : (T is self::TYPE_INT ? int : bool))',
+				new ConditionalTypeNode(
+					new IdentifierTypeNode('T'),
+					new ConstTypeNode(new ConstFetchNode('self', 'TYPE_STRING')),
+					new IdentifierTypeNode('string'),
+					new ConditionalTypeNode(
+						new IdentifierTypeNode('T'),
+						new ConstTypeNode(new ConstFetchNode('self', 'TYPE_INT')),
+						new IdentifierTypeNode('int'),
+						new IdentifierTypeNode('bool'),
+						false
+					),
+					false
+				),
+			],
+			[
+				'(Foo is Bar|Baz ? never : int|string)',
+				new ConditionalTypeNode(
+					new IdentifierTypeNode('Foo'),
+					new UnionTypeNode([
+						new IdentifierTypeNode('Bar'),
+						new IdentifierTypeNode('Baz'),
+					]),
+					new IdentifierTypeNode('never'),
+					new UnionTypeNode([
+						new IdentifierTypeNode('int'),
+						new IdentifierTypeNode('string'),
+					]),
+					false
+				),
+			],
+			[
+				'(' . PHP_EOL .
+				'  TRandList is array ? array<TRandKey, TRandVal> : (' . PHP_EOL .
+				'  TRandList is XIterator ? XIterator<TRandKey, TRandVal> :' . PHP_EOL .
+				'  IteratorIterator<TRandKey, TRandVal>|LimitIterator<TRandKey, TRandVal>' . PHP_EOL .
+				'))',
+				new ConditionalTypeNode(
+					new IdentifierTypeNode('TRandList'),
+					new IdentifierTypeNode('array'),
+					new GenericTypeNode(
+						new IdentifierTypeNode('array'),
+						[
+							new IdentifierTypeNode('TRandKey'),
+							new IdentifierTypeNode('TRandVal'),
+						]
+					),
+					new ConditionalTypeNode(
+						new IdentifierTypeNode('TRandList'),
+						new IdentifierTypeNode('XIterator'),
+						new GenericTypeNode(
+							new IdentifierTypeNode('XIterator'),
+							[
+								new IdentifierTypeNode('TRandKey'),
+								new IdentifierTypeNode('TRandVal'),
+							]
+						),
+						new UnionTypeNode([
+							new GenericTypeNode(
+								new IdentifierTypeNode('IteratorIterator'),
+								[
+									new IdentifierTypeNode('TRandKey'),
+									new IdentifierTypeNode('TRandVal'),
+								]
+							),
+							new GenericTypeNode(
+								new IdentifierTypeNode('LimitIterator'),
+								[
+									new IdentifierTypeNode('TRandKey'),
+									new IdentifierTypeNode('TRandVal'),
+								]
+							),
+						]),
+						false
+					),
+					false
+				),
+			],
+			[
+				'($foo is Bar|Baz ? never : int|string)',
+				new ConditionalTypeForParameterNode(
+					'$foo',
+					new UnionTypeNode([
+						new IdentifierTypeNode('Bar'),
+						new IdentifierTypeNode('Baz'),
+					]),
+					new IdentifierTypeNode('never'),
+					new UnionTypeNode([
+						new IdentifierTypeNode('int'),
+						new IdentifierTypeNode('string'),
+					]),
+					false
+				),
+			],
+			[
+				'(' . PHP_EOL .
+				'  $foo is Bar|Baz' . PHP_EOL .
+				'    ? never' . PHP_EOL .
+				'    : int|string' . PHP_EOL .
+				')',
+				new ConditionalTypeForParameterNode(
+					'$foo',
+					new UnionTypeNode([
+						new IdentifierTypeNode('Bar'),
+						new IdentifierTypeNode('Baz'),
+					]),
+					new IdentifierTypeNode('never'),
+					new UnionTypeNode([
+						new IdentifierTypeNode('int'),
+						new IdentifierTypeNode('string'),
+					]),
+					false
+				),
+			],
+			[
+				'?Currency::CURRENCY_*',
+				new NullableTypeNode(
+					new ConstTypeNode(
+						new ConstFetchNode(
+							'Currency',
+							'CURRENCY_*'
+						)
+					)
+				),
+			],
 		];
 	}
 
